#include <amxmodx>
#include <fakemeta>
#include <fakemeta_util>
#include <hamsandwich>
#include <xs>
#include <zombieplague_x>

#define PLUGIN "[ZP] Extra: CSO Weapon"
#define VERSION "1.0"
#define AUTHOR "KORD_12.7 | qubka"

#pragma ctrlchar '\'

//**********************************************
//* Weapon Settings.                           *
//**********************************************

// Main
#define WEAPON_NAME 				"weapon_m32_zpx"
#define WEAPON_REFERANCE			"weapon_m3"
#define WEAPON_MAX_CLIP				6
#define WEAPON_DEFAULT_AMMO			40
#define WEAPON_MAX_SPEED			1.0

#define WEAPON_TIME_NEXT_IDLE 		5.0
#define WEAPON_TIME_NEXT_ATTACK 	0.58
#define WEAPON_TIME_DELAY_DEPLOY 	1.23
#define WEAPON_TIME_DELAY_RELOAD 	1.25

#define ZP_ITEM_NAME				"M32" 
#define ZP_ITEM_COST				25

#define GRENADE_RADIUS 				300.0
#define GRENADE_DAMAGE 				500.0

// Models
#define MODEL_WORLD				"models/zombie_plague/w_m32.mdl"
#define MODEL_VIEW				"models/zombie_plague/v_m32.mdl"
#define MODEL_PLAYER			"models/zombie_plague/p_m32.mdl"
#define MODEL_GRENADE  	 		"models/zombie_plague/shell_m32.mdl"

#define GRENADE_TRAIL     		"sprites/laserbeam.spr"
#define GRENADE_EXPLOSION  		"sprites/zerogxplode.spr"
#define GRENADE_SMOKE      		"sprites/black_smoke3.spr"

#define SPRITE_MUZZLE			"sprites/zpx/muzzleflash_launcher.spr"

#define MUZZLE_CLASSNAME		"muzzle_m32"
#define MUZZLE_INTOLERANCE		100

// Sounds
#define SOUND_FIRE				"weapons/m32-1.wav"

#define WEAPON_HUD_SPR1			"sprites/zpx/scope_vip_grenade.spr"
#define WEAPON_HUD_SPR2			"sprites/zpx/640hud75.spr"
#define WEAPON_HUD_SPR3			"sprites/zpx/640hud7.spr"
#define WEAPON_HUD_TXT    		"sprites/weapon_m32_zpx.txt"

// Animation
#define ANIM_EXTENSION			"carbine"

// Animation sequences
enum
{	
	ANIM_IDLE,
	ANIM_SHOOT1,
	ANIM_SHOOT2,
	ANIM_INSERT,
	ANIM_AFTER_RELOAD,
	ANIM_START_RELOAD,
	ANIM_DRAW
};

//**********************************************
//* Some macroses.                             *
//**********************************************

#define MDLL_Spawn(%0)			dllfunc(DLLFunc_Spawn, %0)
#define MDLL_Touch(%0,%1)		dllfunc(DLLFunc_Touch, %0, %1)
#define MDLL_USE(%0,%1)			dllfunc(DLLFunc_Use, %0, %1)
#define SET_MODEL(%0,%1)		engfunc(EngFunc_SetModel, %0, %1)
#define SET_ORIGIN(%0,%1)		engfunc(EngFunc_SetOrigin, %0, %1)

#define PRECACHE_MODEL(%0)		engfunc(EngFunc_PrecacheModel, %0)
#define PRECACHE_SOUND(%0)		engfunc(EngFunc_PrecacheSound, %0)
#define PRECACHE_GENERIC(%0)		engfunc(EngFunc_PrecacheGeneric, %0)

#define MESSAGE_BEGIN(%0,%1,%2,%3)	engfunc(EngFunc_MessageBegin, %0, %1, %2, %3)
#define MESSAGE_END()			message_end()

#define WRITE_ANGLE(%0)			engfunc(EngFunc_WriteAngle, %0)
#define WRITE_BYTE(%0)			write_byte(%0)
#define WRITE_COORD(%0)			engfunc(EngFunc_WriteCoord, %0)
#define WRITE_STRING(%0)		write_string(%0)
#define WRITE_SHORT(%0)			write_short(%0)

#define NUMBER_OF_ENTITIES() engfunc(EngFunc_NumberOfEntities)

#define INSTANCE(%0)			((%0 == -1) ? 0 : %0)

#define BitSet(%0,%1) (%0 |= (1 << (%1 - 1)))
#define BitClear(%0,%1) (%0 &= ~(1 << (%1 - 1)))
#define BitCheck(%0,%1) (%0 & (1 << (%1 - 1)))

//**********************************************
//* PvData Offsets.                            *
//**********************************************

// Linux extra offsets
#define extra_offset_weapon		4
#define extra_offset_player		5

new g_bitIsConnected;
#define m_rgpPlayerItems_CWeaponBox	34

// CBasePlayerItem
#define m_pPlayer			41
#define m_pNext				42

// CSprite
#define m_maxFrame 				 35

// CBasePlayerWeapon
#define m_flNextPrimaryAttack		46
#define m_flNextSecondaryAttack		47
#define m_flTimeWeaponIdle		48
#define m_iPrimaryAmmoType		49
#define m_iClip				51
#define m_fInReload			54
#define m_fInSpecialReload	55
#define m_iDirection			60
#define m_iShotsFired			64
#define m_fWeaponState			74
#define m_flLastEventCheck  	38
#define m_iLastZoom 			109
#define OFFSET_M3_AMMO 			381

// CBaseMonster
#define m_flNextAttack			83

// CBasePlayer
#define m_iFOV				363
#define m_fResumeZoom       110
#define m_rgpPlayerItems_CBasePlayer	367
#define m_pActiveItem			373
#define m_rgAmmo_CBasePlayer		376
#define m_szAnimExtention		492

#define IsValidPev(%0) (pev_valid(%0) == 2)

#define GRENADE_CLASS "m32_grenade"

#define IsObserver(%0) pev(%0,pev_iuser1)

#define OBS_IN_EYE 4

#define HIDEHUD_CROSSHAIR (1 << 6)
#define HIDEHUD_SPECTORHAIR (1 << 7)

//**********************************************
//* Let's code our weapon.                     *
//**********************************************

new sTrail; 
new sExplo; 
new sSmoke;

Weapon_OnPrecache()
{
	PRECACHE_MODEL(MODEL_VIEW);
	PRECACHE_MODEL(MODEL_PLAYER);
	PRECACHE_MODEL(MODEL_WORLD) ;
	PRECACHE_MODEL(MODEL_GRENADE);
	
	PRECACHE_SOUNDS_FROM_MODEL(MODEL_VIEW);

	PRECACHE_SOUND(SOUND_FIRE);
	
	PRECACHE_GENERIC(WEAPON_HUD_TXT);
	PRECACHE_GENERIC(WEAPON_HUD_SPR1);
	PRECACHE_GENERIC(WEAPON_HUD_SPR2);
	PRECACHE_GENERIC(WEAPON_HUD_SPR3);
	
	sTrail = PRECACHE_MODEL(GRENADE_TRAIL);
	sExplo = PRECACHE_MODEL(GRENADE_EXPLOSION);
	sSmoke = PRECACHE_MODEL(GRENADE_SMOKE);
}

Weapon_OnSpawn(const iItem)
{
	// Setting world model.
	SET_MODEL(iItem, MODEL_WORLD);
}

Weapon_OnDeploy(const iItem, const iPlayer, const iClip, const iAmmoPrimary)
{
	#pragma unused iClip, iAmmoPrimary
	
	static iszViewModel;
	if (iszViewModel || (iszViewModel = engfunc(EngFunc_AllocString, MODEL_VIEW)))
	{
		set_pev_string(iPlayer, pev_viewmodel2, iszViewModel);
	}
	
	static iszPlayerModel;
	if (iszPlayerModel || (iszPlayerModel = engfunc(EngFunc_AllocString, MODEL_PLAYER)))
	{
		set_pev_string(iPlayer, pev_weaponmodel2, iszPlayerModel);
	}
	
	//Hide Radar, Hp, Armor, CrossHair
	message_begin(MSG_ONE, get_user_msgid("HideWeapon"), _, iPlayer);
	write_byte(HIDEHUD_CROSSHAIR | HIDEHUD_SPECTORHAIR);
	message_end();
	
	set_pdata_int(iItem, m_fInSpecialReload, 0, extra_offset_weapon);
	
	set_pdata_string(iPlayer, m_szAnimExtention * 4, ANIM_EXTENSION, -1, extra_offset_player * 4);
	
	set_pdata_float(iItem, m_flTimeWeaponIdle, WEAPON_TIME_DELAY_DEPLOY, extra_offset_weapon);
	set_pdata_float(iPlayer, m_flNextAttack, WEAPON_TIME_DELAY_DEPLOY, extra_offset_player);
	
	// Deploy with default weapon settings
	Weapon_DefaultDeploy(iItem, iPlayer, MODEL_VIEW, MODEL_PLAYER, ANIM_DRAW, ANIM_EXTENSION);
}

Weapon_OnHolster(const iItem, const iPlayer, const iClip, const iAmmoPrimary)
{
	#pragma unused iPlayer, iClip, iAmmoPrimary
	
	//Show Radar, Hp, Armor, CrossHair
	message_begin(MSG_ONE, get_user_msgid("HideWeapon"), _, iPlayer);
	write_byte(1>>6 | 1>>7);
	message_end();
	
	// Cancel any reload in progress.
	set_pdata_int(iItem, m_fInSpecialReload, 0, extra_offset_weapon);
}

Weapon_OnIdle(const iItem, const iPlayer, const iClip, const iAmmoPrimary)
{
	#pragma unused iClip, iAmmoPrimary
	
	ExecuteHamB(Ham_Weapon_ResetEmptySound, iItem);

	static iReloadMode; 
	iReloadMode = get_pdata_int(iItem, m_fInSpecialReload, extra_offset_weapon);
	
	if(iClip == WEAPON_MAX_CLIP)
	{
		if(iReloadMode == 2)
		{
			Weapon_ReloadEnd(iItem, iPlayer);
			return;
		}
	}
	
	if(iAmmoPrimary <= 0)
	{
		if(iReloadMode == 2)
		{
			Weapon_ReloadEnd(iItem, iPlayer);
			return;
		}
	}

	switch(iReloadMode)
	{
		case 1:
		{		
			if (get_pdata_int(iItem, m_flTimeWeaponIdle, extra_offset_weapon) > 0.0)
			{
				return;
			}
			
			Weapon_SendAnim(iPlayer, ANIM_INSERT);
				
			set_pdata_int(iItem, m_fInSpecialReload, 2, extra_offset_weapon);
			set_pdata_float(iItem, m_flTimeWeaponIdle, 0.7, extra_offset_weapon);
			
			static szAnimation[64];

			formatex(szAnimation, charsmax(szAnimation), "ref_reload_shotgun");
			Player_SetAnimation(iPlayer, szAnimation);
			
		}
		case 2:
		{
			set_pdata_int(iItem, m_iClip, iClip + 1, extra_offset_weapon);
			set_pdata_int(iPlayer, OFFSET_M3_AMMO, iAmmoPrimary-1, extra_offset_player);
			set_pdata_int(iItem, m_fInSpecialReload, 1, extra_offset_weapon);
		}
	}
	
	if(!iReloadMode)
	{
		if (get_pdata_int(iItem, m_flTimeWeaponIdle, extra_offset_weapon) > 0.0)
		{
			return;
		}
	
		set_pdata_float(iItem, m_flTimeWeaponIdle, WEAPON_TIME_NEXT_IDLE, extra_offset_weapon);
		Weapon_SendAnim(iPlayer, ANIM_IDLE);
	}
}

Weapon_OnReload(const iItem, const iPlayer, const iClip, const iAmmoPrimary)
{
	#pragma unused iClip

	if(iAmmoPrimary <= 0 )
	{
		return HAM_IGNORED;
	}
	
	if(iClip >= WEAPON_MAX_CLIP)
	{
		return HAM_IGNORED;
	}
	
	static iReloadMode; 
	iReloadMode = get_pdata_int(iItem, m_fInSpecialReload, extra_offset_weapon);
	
	if (get_pdata_int(iPlayer, m_iFOV, extra_offset_player) != 89)
	{
		Weapon_OnSecondaryAttack(iItem, iPlayer, iClip, iAmmoPrimary);
	}

	if(iClip <= 0) 
	{
		switch(iReloadMode)
		{
			case 0:
			{
				Weapon_SendAnim(iPlayer, ANIM_START_RELOAD);
			
				set_pdata_float(iItem, m_flNextPrimaryAttack, 0.9, extra_offset_weapon);
				set_pdata_float(iItem, m_flTimeWeaponIdle, 0.9, extra_offset_weapon);
			
				set_pdata_int(iItem, m_fInSpecialReload, 1, extra_offset_weapon);
			
				return HAM_IGNORED;
			}
			case 1:
			{		
				if (get_pdata_int(iItem, m_flTimeWeaponIdle, extra_offset_weapon) > 0.0)
				{
					return HAM_IGNORED;
				}
				
				Weapon_SendAnim(iPlayer, ANIM_INSERT);
					
				set_pdata_int(iItem, m_fInSpecialReload, 2, extra_offset_weapon);
				set_pdata_float(iItem, m_flTimeWeaponIdle, 0.7, extra_offset_weapon);
				
				static szAnimation[64];

				formatex(szAnimation, charsmax(szAnimation), "ref_reload_shotgun");
				Player_SetAnimation(iPlayer, szAnimation);
				
			}
			case 2:
			{
				set_pdata_int(iItem, m_iClip, iClip + 1, extra_offset_weapon);
				set_pdata_int(iPlayer, OFFSET_M3_AMMO, iAmmoPrimary-1, extra_offset_player);
				set_pdata_int(iItem, m_fInSpecialReload, 1, extra_offset_weapon);
			}
		}
		return HAM_IGNORED;
	}
	
	switch(iReloadMode)
	{
		case 0:
		{
			Weapon_SendAnim(iPlayer, ANIM_START_RELOAD);
			
			set_pdata_float(iItem, m_flNextPrimaryAttack, 0.9, extra_offset_weapon);
			set_pdata_float(iItem, m_flTimeWeaponIdle, 0.9, extra_offset_weapon);
			
			set_pdata_int(iItem, m_fInSpecialReload, 1, extra_offset_weapon);
			
			return HAM_IGNORED;
		}
	}
	return HAM_IGNORED;
}

// Reload Ended
Weapon_ReloadEnd(const iItem, const iPlayer)
{
	Weapon_SendAnim(iPlayer, ANIM_AFTER_RELOAD);
	
	set_pdata_float(iItem, m_flNextPrimaryAttack, 1.2, extra_offset_weapon);
	set_pdata_float(iItem, m_flTimeWeaponIdle, 1.2, extra_offset_weapon);
	
	set_pdata_int(iItem, m_fInSpecialReload, 0, extra_offset_weapon);
}

Weapon_OnSecondaryAttack(const iItem, const iPlayer, const iClip, const iAmmoPrimary)
{
	#pragma unused iClip, iAmmoPrimary

	set_pdata_float(iItem, m_flNextSecondaryAttack, 0.3, extra_offset_weapon);
	set_pdata_int(iPlayer, m_iFOV, get_pdata_int(iPlayer, m_iFOV, extra_offset_player) == 89 ? 55 : 89, extra_offset_player);
}

//********************************************************************
//*           Don't modify the code below this line unless            *
//*          	 you know _exactly_ what you are doing!!!             *
//*********************************************************************

#define MSGID_WEAPONLIST 78

new g_iszWeaponKey;
new g_iszMuzzleKey;

new g_iItemID;

#define IsCustomItem(%0) (pev(%0, pev_impulse) == g_iszWeaponKey)
#define IsCustomMuzzle(%0) (pev(%0, pev_impulse) == g_iszMuzzleKey)

public plugin_precache()
{
	Weapon_OnPrecache();
	
	g_iszWeaponKey = engfunc(EngFunc_AllocString, WEAPON_NAME);
	g_iszMuzzleKey = engfunc(EngFunc_AllocString, MUZZLE_CLASSNAME);
	
	register_message(MSGID_WEAPONLIST, "MsgHook_WeaponList");
}

public plugin_init()
{
	register_plugin(PLUGIN, VERSION, AUTHOR);
	
	g_iItemID = zp_register_extra_item("Milkor-32", 0, 0);
	
	register_forward(FM_SetModel,		"FakeMeta_SetModel",			false);
	register_forward(FM_UpdateClientData,	"FakeMeta_UpdateClientData_Post",	true);
	
	register_event("HLTV", "Round_Start", "a", "1=0", "2=0");
	
	// Weaponbox
	RegisterHam(Ham_Spawn, 		"weaponbox", 		"HamHook_Weaponbox_Spawn_Post", true);
	
	// Item (weapon) hooks
	RegisterHam(Ham_Item_Deploy,		WEAPON_REFERANCE, "HamHook_Item_Deploy_Post",	true);
	RegisterHam(Ham_Item_Holster,		WEAPON_REFERANCE, "HamHook_Item_Holster",	false);
	RegisterHam(Ham_Item_AddToPlayer,	WEAPON_REFERANCE, "HamHook_Item_AddToPlayer",	false);
	RegisterHam(Ham_Weapon_WeaponIdle,	WEAPON_REFERANCE, "HamHook_Item_WeaponIdle",	false);
	RegisterHam(Ham_Weapon_PrimaryAttack,	WEAPON_REFERANCE, "HamHook_Item_PrimaryAttack_Pre",	false);
	RegisterHam(Ham_Weapon_PrimaryAttack,	WEAPON_REFERANCE, "HamHook_Item_PrimaryAttack_Post",	false);
	RegisterHam(Ham_Weapon_Reload,		WEAPON_REFERANCE, "HamHook_Item_Reload",	false);
	RegisterHam(Ham_Item_PostFrame,		WEAPON_REFERANCE, "HamHook_Item_PostFrame",	false);
	
	register_forward(FM_Touch , "FakeMeta_Touch",						false);
	register_forward(FM_PlaybackEvent,	"FakeMeta_PlaybackEvent",		false);
	
	// Muzzle flash think
	RegisterHam(Ham_Think, "env_sprite",	"MuzzleFlash_Think", false);
}

public Round_Start()
{
	new ent = -1;
	while ((ent = fm_find_ent_by_class(ent, GRENADE_CLASS)))
	{
		set_pev(ent, pev_flags, pev(ent, pev_flags) | FL_KILLME);
	}
}

public zp_extra_item_selected(id, itemid)
{
	if (itemid == g_iItemID)
	{
		Weapon_Give(id);
	}
}

//**********************************************
//* Block client weapon.                       *
//**********************************************

public FakeMeta_UpdateClientData_Post(const iPlayer, const iSendWeapons, const CD_Handle)
{
	static iActiveItem;iActiveItem = get_pdata_cbase(iPlayer, m_pActiveItem, extra_offset_player);
	
	if (!IsValidPev(iActiveItem) || !IsCustomItem(iActiveItem))
	{
		return FMRES_IGNORED;
	}

	set_cd(CD_Handle, CD_flNextAttack, get_gametime() + 0.001);
	return FMRES_IGNORED;
}

public FakeMeta_PlaybackEvent() 
{
	return FMRES_SUPERCEDE;
}

//**********************************************
//* Item (weapon) hooks.                       *
//**********************************************

	#define _call.%0(%1,%2) \
									\
	Weapon_On%0							\
	(								\
		%1, 							\
		%2,							\
									\
		get_pdata_int(%1, m_iClip, extra_offset_weapon),	\
									\
		GetAmmoInventory(%2, PrimaryAmmoIndex(%1))		\
	) 


public HamHook_Item_Deploy_Post(const iItem)
{
	new iPlayer; 
	
	if (!CheckItem(iItem, iPlayer))
	{
		return HAM_IGNORED;
	}
	
	_call.Deploy(iItem, iPlayer);
	return HAM_IGNORED;
}

public HamHook_Item_Holster(const iItem)
{
	new iPlayer; 
	
	if (!CheckItem(iItem, iPlayer))
	{
		return HAM_IGNORED;
	}
	
	set_pev(iPlayer, pev_viewmodel, 0);
	set_pev(iPlayer, pev_weaponmodel, 0);
	
	_call.Holster(iItem, iPlayer);
	return HAM_SUPERCEDE;
}

public HamHook_Item_WeaponIdle(const iItem)
{
	static iPlayer; 
	
	if (!CheckItem(iItem, iPlayer))
	{
		return HAM_IGNORED;
	}

	_call.Idle(iItem, iPlayer);
	return HAM_SUPERCEDE;
}

public HamHook_Item_PrimaryAttack_Pre(const iItem)
{
	static iPlayer;
	
	if (CheckItem(iItem, iPlayer))
	{
		return HAM_SUPERCEDE;
	}
	
	return HAM_IGNORED;
}

public HamHook_Item_PrimaryAttack_Post(const iItem)
{
	static iPlayer;
	static iReloadMode; 
	static iClip; 
	
	if (!CheckItem(iItem, iPlayer))
	{
		return HAM_IGNORED;
	}
	
	iClip = get_pdata_int(iItem, m_iClip, extra_offset_weapon);
	iReloadMode = get_pdata_int(iItem, m_fInSpecialReload, extra_offset_weapon);
	
	if(iReloadMode > 0 && iClip > 0)
	{
		Weapon_ReloadEnd(iItem, iPlayer);
		return HAM_IGNORED;
	}
	
	if (iClip <= 0  || pev(iPlayer, pev_waterlevel) == 3)
	{
		return HAM_IGNORED;
	}

	static szAnimation[64];

	formatex(szAnimation, charsmax(szAnimation), "ref_shoot_notmuzl");
	Player_SetAnimation(iPlayer, szAnimation);
	
	set_pev(iPlayer, pev_punchangle, Float:{ -5.0, 0.0, 0.0 });
	
	// Fixed Shell Eject
	set_pdata_float(iPlayer, 111, 99999999.0, extra_offset_player);
	
	//cs_set_weapon_ammo(iItem, cs_get_weapon_ammo(iItem) - 1);
	set_pdata_int(iItem, m_iClip, iClip - 1, extra_offset_weapon);
	
	// Draw muzzleflash
	//Weapon_MuzzleFlash(iPlayer, SPRITE_MUZZLE, 0.08, 13.0, Float: {121.0, 150.0, 150.0}, 255.0);  
	
	Fire_Grenade(iPlayer);
	
	emit_sound(iPlayer, CHAN_WEAPON, SOUND_FIRE, VOL_NORM, ATTN_NORM, 0, PITCH_NORM);
	
	Weapon_SendAnim(iPlayer, random_num(ANIM_SHOOT1, ANIM_SHOOT2));
	
	set_pdata_float(iItem, m_flNextPrimaryAttack , WEAPON_TIME_NEXT_ATTACK , extra_offset_weapon);
	set_pdata_float(iItem, m_flTimeWeaponIdle, WEAPON_TIME_NEXT_ATTACK + 3.0  , extra_offset_weapon);
	
	return HAM_SUPERCEDE;
}

public Fire_Grenade(const iPlayer)
{
	if (global_get(glb_maxEntities) - engfunc(EngFunc_NumberOfEntities) < 100)
	{
		return;
	}

	if(is_user_alive(iPlayer))
	{
		new pEntity;
 
		static iszAllocStringCached;
		if (iszAllocStringCached || (iszAllocStringCached = engfunc(EngFunc_AllocString, "info_target")))
		{
			   pEntity = engfunc(EngFunc_CreateNamedEntity, iszAllocStringCached);
		}
			
		if(pev_valid(pEntity))
		{			
			new Float:Origin[3], Float:Angles[3], Float:Velocity[3];
			
			engfunc(EngFunc_GetAttachment, iPlayer, 0, Origin, Angles);
			pev(iPlayer, pev_angles, Angles);
	
			set_pev(pEntity, pev_origin, Origin);
			set_pev(pEntity, pev_angles, Angles);
			set_pev(pEntity, pev_v_angle, Angles);
			set_pev(pEntity, pev_solid, 2);
			set_pev(pEntity, pev_movetype, MOVETYPE_TOSS);
			set_pev(pEntity, pev_classname, GRENADE_CLASS);
			set_pev(pEntity, pev_owner, iPlayer);
			engfunc(EngFunc_SetModel, pEntity, MODEL_GRENADE);
			
			velocity_by_aim(iPlayer, 1500, Velocity);
			set_pev(pEntity, pev_velocity, Velocity);
			
			engfunc(EngFunc_VecToAngles, Velocity, Angles);
			
			set_pev(pEntity, pev_angles, Angles); 
			set_pev(pEntity, pev_v_angle, Angles) ;

			static Classname[32];
			pev(pEntity, pev_classname, Classname, sizeof(Classname));
			
			if(equal(Classname, GRENADE_CLASS))
			{
				// Launch sound
				message_begin(MSG_BROADCAST, SVC_TEMPENTITY);
				write_byte(TE_BEAMFOLLOW); 
				write_short(pEntity); 
				write_short(sTrail); 
				write_byte(10); 
				write_byte(4); 
				write_byte(255); 
				write_byte(255);
				write_byte(255);
				write_byte(50); 
				message_end();
			}
		}
	}
}

public HamHook_Item_Reload(const iItem)
{
	static iPlayer; 
	
	if (!CheckItem(iItem, iPlayer))
	{
		return HAM_IGNORED;
	}
	
	_call.Reload(iItem, iPlayer);
	return HAM_SUPERCEDE;
}

public HamHook_Item_PostFrame(const iItem)
{
	static iButton, iPlayer; 
	
	if (!CheckItem(iItem, iPlayer))
	{
		return HAM_IGNORED;
	}
	
	// Call secondary attack
	if ((iButton = pev(iPlayer, pev_button)) & IN_ATTACK2 
		&& get_pdata_float(iItem, m_flNextSecondaryAttack, extra_offset_weapon) <= 0.0)
	{
		_call.SecondaryAttack(iItem, iPlayer);
		set_pev(iPlayer, pev_button, iButton & ~IN_ATTACK2);
	}
	
	return HAM_IGNORED;
}

//**********************************************
//* Muzzleflash stuff.                *
//**********************************************

#define Sprite_SetScale(%0,%1) set_pev(%0, pev_scale, %1)
#define Sprite_SetFramerate(%0,%1) set_pev(%0, pev_framerate, %1)

stock Sprite_SetTransparency(const iSprite, const iRendermode, const Float: vecColor[3], const Float: flAmt, const iFx = kRenderFxNone)
{
	set_pev(iSprite, pev_rendermode, iRendermode);
	set_pev(iSprite, pev_rendercolor, vecColor);
	set_pev(iSprite, pev_renderamt, flAmt);
	set_pev(iSprite, pev_renderfx, iFx);
}

#define NUMBER_OF_ENTITIES() engfunc(EngFunc_NumberOfEntities)

stock Weapon_MuzzleFlash(const iPlayer, const szMuzzleSprite[], const Float: flScale, const Float: flFramerate, const Float: vecColor[3], const Float: flBrightness)
{
	if (global_get(glb_maxEntities) - NUMBER_OF_ENTITIES() < MUZZLE_INTOLERANCE)
	{
		return FM_NULLENT;
	}
	
	static iSprite, iszAllocStringCached;
	if (iszAllocStringCached || (iszAllocStringCached = engfunc(EngFunc_AllocString, "env_sprite")))
	{
		iSprite = engfunc(EngFunc_CreateNamedEntity, iszAllocStringCached);
	}
	
	if (!IsValidPev(iSprite))
	{
		return FM_NULLENT;
	}
	
	set_pev(iSprite, pev_model, szMuzzleSprite);
	set_pev(iSprite, pev_spawnflags, SF_SPRITE_ONCE);
	
	set_pev(iSprite, pev_classname, MUZZLE_CLASSNAME);
	set_pev(iSprite, pev_impulse, g_iszMuzzleKey);
	set_pev(iSprite, pev_owner, iPlayer);
	
	set_pev(iSprite, pev_aiment, iPlayer);
	//set_pev(iSprite, pev_body, 1);
	
	Sprite_SetTransparency(iSprite, kRenderTransAdd, vecColor, flBrightness);
	Sprite_SetFramerate(iSprite, flFramerate);
	Sprite_SetScale(iSprite, flScale);
	
	MDLL_Spawn(iSprite);

	return iSprite;
}

public MuzzleFlash_Think(const iSprite)
{
	static Float: flFrame;
	
	if (!IsValidPev(iSprite) || !IsCustomMuzzle(iSprite))
	{
		return HAM_IGNORED;
	}
	
	if (pev(iSprite, pev_frame, flFrame) && ++flFrame - 1.0 < get_pdata_float(iSprite, m_maxFrame, extra_offset_weapon))
	{
		set_pev(iSprite, pev_frame, flFrame);
		set_pev(iSprite, pev_nextthink, get_gametime() + 0.2);
		
		return HAM_SUPERCEDE;
	}

	set_pev(iSprite, pev_flags, FL_KILLME);
	return HAM_SUPERCEDE;
}

/*public AddToFullPack_Post(const es, const e, const iEntity, const iHost, const hostflags, const IsPlayer, const pSet)
{
	if (!IsPlayer && IsValidPev(iEntity) && IsValidPev(iHost) 
		&& IsObserver(iHost) == OBS_IN_EYE && pev(iEntity, pev_owner) == pev(iHost, pev_iuser2))
	{
		if (IsCustomMuzzle(iEntity))
		{
			set_es(es, ES_Skin, iHost);
			set_es(es, ES_AimEnt, iHost);
		}
	}
	
	return FMRES_IGNORED;
}*/

//**********************************************
//* PRECACHE.                              *
//**********************************************

PRECACHE_SOUNDS_FROM_MODEL(const szModelPath[])
{
	new iFile;
	
	if ((iFile = fopen(szModelPath, "rt")))
	{
		new szSoundPath[64];
		
		new iNumSeq, iSeqIndex;
		new iEvent, iNumEvents, iEventIndex;
		
		fseek(iFile, 164, SEEK_SET);
		fread(iFile, iNumSeq, BLOCK_INT);
		fread(iFile, iSeqIndex, BLOCK_INT);
		
		for (new k, i = 0; i < iNumSeq; i++)
		{
			fseek(iFile, iSeqIndex + 48 + 176 * i, SEEK_SET);
			fread(iFile, iNumEvents, BLOCK_INT);
			fread(iFile, iEventIndex, BLOCK_INT);
			fseek(iFile, iEventIndex + 176 * i, SEEK_SET);

			for (k = 0; k < iNumEvents; k++)
			{
				fseek(iFile, iEventIndex + 4 + 76 * k, SEEK_SET);
				fread(iFile, iEvent, BLOCK_INT);
				fseek(iFile, 4, SEEK_CUR);
				
				if (iEvent != 5004)
				{
					continue;
				}

				fread_blocks(iFile, szSoundPath, 64, BLOCK_CHAR);
				
				if (strlen(szSoundPath))
				{
					strtolower(szSoundPath);
					PRECACHE_SOUND(szSoundPath);
				}
			}
		}
	}
	
	fclose(iFile);
}

//**********************************************
//* Create and check our custom weapon.        *
//**********************************************

Weapon_Create(const Float: vecOrigin[3] = {0.0, 0.0, 0.0}, const Float: vecAngles[3] = {0.0, 0.0, 0.0})
{
	new iWeapon;

	static iszAllocStringCached;
	if (iszAllocStringCached || (iszAllocStringCached = engfunc(EngFunc_AllocString, WEAPON_REFERANCE)))
	{
		iWeapon = engfunc(EngFunc_CreateNamedEntity, iszAllocStringCached);
	}
	
	if (!IsValidPev(iWeapon))
	{
		return FM_NULLENT;
	}
	
	MDLL_Spawn(iWeapon);
	SET_ORIGIN(iWeapon, vecOrigin);
	
	set_pdata_int(iWeapon, m_iClip, WEAPON_MAX_CLIP, extra_offset_weapon);

	set_pev_string(iWeapon, pev_classname, g_iszWeaponKey );
	set_pev(iWeapon, pev_impulse, g_iszWeaponKey);
	set_pev(iWeapon, pev_angles, vecAngles);
	
	Weapon_OnSpawn(iWeapon);
	
	return iWeapon;
}

Weapon_Give(const iPlayer)
{
	if (!IsValidPev(iPlayer))
	{
		return FM_NULLENT;
	}
	
	new iWeapon, Float: vecOrigin[3];
	pev(iPlayer, pev_origin, vecOrigin);
	
	if ((iWeapon = Weapon_Create(vecOrigin)) != FM_NULLENT)
	{
		Player_DropWeapons(iPlayer, ExecuteHamB(Ham_Item_ItemSlot, iWeapon));
		
		set_pev(iWeapon, pev_spawnflags, pev(iWeapon, pev_spawnflags) | SF_NORESPAWN);
		MDLL_Touch(iWeapon, iPlayer);
		
		SetAmmoInventory(iPlayer, PrimaryAmmoIndex(iWeapon), WEAPON_DEFAULT_AMMO);
		
		return iWeapon;
	}
	
	return FM_NULLENT;
}

Player_DropWeapons(const iPlayer, const iSlot)
{
	new szWeaponName[32], iItem = get_pdata_cbase(iPlayer, m_rgpPlayerItems_CBasePlayer + iSlot, extra_offset_player);

	while (IsValidPev(iItem))
	{
		pev(iItem, pev_classname, szWeaponName, charsmax(szWeaponName));
		engclient_cmd(iPlayer, "drop", szWeaponName);

		iItem = get_pdata_cbase(iItem, m_pNext, extra_offset_weapon);
	}
}

Weapon_SendAnim(const iPlayer, const iAnim)
{
	set_pev(iPlayer, pev_weaponanim, iAnim);

	message_begin(MSG_ONE_UNRELIABLE, SVC_WEAPONANIM, {0.0, 0.0, 0.0}, iPlayer);
	write_byte(iAnim);
	write_byte(0);
	message_end();
}

stock Weapon_DefaultDeploy(const iItem, const iPlayer, const szViewModel[], const szWeaponModel[], const iAnim, const szAnimExt[])
{
	#pragma unused iItem

	set_pev(iPlayer, pev_viewmodel2, szViewModel);
	set_pev(iPlayer, pev_weaponmodel2, szWeaponModel);
	set_pev(iPlayer, pev_fov, 89.0);
	
	set_pdata_int(iPlayer, m_iFOV, 89, extra_offset_player);
	set_pdata_int(iPlayer, m_fResumeZoom, 0, extra_offset_player);
	set_pdata_int(iPlayer, m_iLastZoom, 89, extra_offset_player);
	
	set_pdata_string(iPlayer, m_szAnimExtention * 4, szAnimExt, -1, extra_offset_player * 4);
	
	//this->m_flFamasBurstSpread = *(float *)gpGlobals;
	
	// Delay for blocking UpdateClientData
	set_pdata_float(iItem, m_flLastEventCheck, get_gametime() + 0.2, extra_offset_weapon);

	Weapon_SendAnim(iPlayer, iAnim);
}



#define BitSet(%0,%1) (%0 |= (1 << (%1 - 1)))
#define BitClear(%0,%1) (%0 &= ~(1 << (%1 - 1)))
#define BitCheck(%0,%1) (%0 & (1 << (%1 - 1)))

public client_putinserver(id)
{
	BitSet(g_bitIsConnected, id);
}

public client_disconnected(id)
{
	BitClear(g_bitIsConnected, id);
}


bool: CheckItem(const iItem, &iPlayer)
{
	if (!IsValidPev(iItem) || !IsCustomItem(iItem))
	{
		return false;
	}
	
	iPlayer = get_pdata_cbase(iItem, m_pPlayer, extra_offset_weapon);
	
	if (!IsValidPev(iPlayer) || !BitCheck(g_bitIsConnected, iPlayer))
	{
		return false;
	}
	
	return true;
}

//**********************************************
//* Weapon effect.                             *
//**********************************************

public FakeMeta_Touch(pEntity, i_Other)
{
	if (!pev_valid(pEntity))
	{
		return FMRES_IGNORED ;
	}
	
	static classname[32];
	pev(pEntity, pev_classname, classname, 31);

	if (equal(classname, GRENADE_CLASS))
	{
		// Get it's origin
		new Float:originF[3];
		pev(pEntity, pev_origin, originF);
		
		static iAttacker; iAttacker = pev(pEntity, pev_owner);
			
		if(!is_user_connected(iAttacker))
		{
			// Destroy ent
			set_pev(pEntity, pev_flags, pev(pEntity, pev_flags) | FL_KILLME);
			return FMRES_IGNORED;
		}
		
		engfunc(EngFunc_MessageBegin, MSG_PAS, SVC_TEMPENTITY, originF, 0);
		write_byte(TE_WORLDDECAL);
		engfunc(EngFunc_WriteCoord, originF[0]);
		engfunc(EngFunc_WriteCoord, originF[1]);
		engfunc(EngFunc_WriteCoord, originF[2]);
		write_byte(engfunc(EngFunc_DecalIndex,"{scorch3"));
		message_end();
		
		message_begin(MSG_ALL, SVC_TEMPENTITY);
		write_byte(TE_SPARKS);
		engfunc(EngFunc_WriteCoord, originF[0]);
		engfunc(EngFunc_WriteCoord, originF[1]);
		engfunc(EngFunc_WriteCoord, originF[2]+30.0);
		message_end();
		
		message_begin(MSG_ALL, SVC_TEMPENTITY);
		write_byte(TE_SPARKS);
		engfunc(EngFunc_WriteCoord, originF[0]+30.0);
		engfunc(EngFunc_WriteCoord, originF[1]);
		engfunc(EngFunc_WriteCoord, originF[2]+30.0);
		message_end();
		
		message_begin(MSG_ALL, SVC_TEMPENTITY);
		write_byte(TE_SPARKS);
		engfunc(EngFunc_WriteCoord, originF[0]);
		engfunc(EngFunc_WriteCoord, originF[1]+30.0);
		engfunc(EngFunc_WriteCoord, originF[2]+30.0);
		message_end();
		
		message_begin(MSG_ALL, SVC_TEMPENTITY);
		write_byte(TE_SPARKS);
		engfunc(EngFunc_WriteCoord, originF[0]);
		engfunc(EngFunc_WriteCoord, originF[1]+40.0);
		engfunc(EngFunc_WriteCoord, originF[2]+40.0);
		message_end();
		
		message_begin(MSG_ALL, SVC_TEMPENTITY);
		write_byte(TE_SPARKS);
		engfunc(EngFunc_WriteCoord, originF[0]);
		engfunc(EngFunc_WriteCoord, originF[1]+40.0);
		engfunc(EngFunc_WriteCoord, originF[2]+30.0);
		message_end();
		
		message_begin(MSG_ALL, SVC_TEMPENTITY);
		write_byte(TE_SPARKS);
		engfunc(EngFunc_WriteCoord, originF[0]);
		engfunc(EngFunc_WriteCoord, originF[1]+30.0);
		engfunc(EngFunc_WriteCoord, originF[2]+40.0);
		message_end();
		
		// Draw explosion
		message_begin(MSG_BROADCAST, SVC_TEMPENTITY);
		write_byte(TE_EXPLOSION);
		engfunc(EngFunc_WriteCoord, originF[0]);
		engfunc(EngFunc_WriteCoord, originF[1]);
		engfunc(EngFunc_WriteCoord, originF[2]+30.0);
		write_short(sExplo);
		write_byte(30);
		write_byte(30);
		write_byte(0);
		message_end();
			
		// Draw smoke
		message_begin(MSG_BROADCAST, SVC_TEMPENTITY);
		write_byte(TE_SMOKE);
		engfunc(EngFunc_WriteCoord, originF[0]);
		engfunc(EngFunc_WriteCoord, originF[1]);
		engfunc(EngFunc_WriteCoord, originF[2]+40.0);
		write_short(sSmoke);
		write_byte(15);
		write_byte(30);
		message_end();
		
		if(pev_valid(i_Other) && !is_user_connected(i_Other))
		{
			static szClassname[32];
			pev(i_Other, pev_classname, szClassname, 31);
			if(equal(szClassname, "func_breakable") && (pev(i_Other, pev_solid) != SOLID_NOT))
			{
				MDLL_USE(i_Other, pEntity);
			}
		}
		
		// Get owner
		new iOwner = pev(pEntity, pev_owner);	
		new pNull = FM_NULLENT;	

		while((pNull = fm_find_ent_in_sphere(pNull, originF, GRENADE_RADIUS)) != 0)
		{
			if(IsValidPev(pNull) && pev(pNull, pev_takedamage) != DAMAGE_NO && pev(pNull, pev_solid) != SOLID_NOT)
			{
				if(is_user_connected(pNull)  && (zp_get_user_zombie(pNull) || zp_get_user_nemesis(pNull)))
				{
					static Float:originV[3], Float:dist, Float:damage;
					pev(pNull, pev_origin, originV);
					dist = get_distance_f(originF, originV);
					damage = GRENADE_DAMAGE - (GRENADE_DAMAGE/GRENADE_RADIUS) * dist;
					if(damage > 0.0)
					{
						ExecuteHamB(Ham_TakeDamage, pNull, iOwner, iOwner, damage, DMG_BULLET);
					}
				}
			}
		}
		
		set_pev(pEntity, pev_flags, pev(pEntity, pev_flags) | FL_KILLME);
	}
	return FMRES_IGNORED;
}

//**********************************************
//* Weapon list update.                        *
//**********************************************

public HamHook_Item_AddToPlayer(const iItem, const iPlayer)
{
	if (!IsValidPev(iItem) || !IsValidPev(iPlayer))
	{
		return HAM_IGNORED;
	}
	
	MsgHook_WeaponList(MSGID_WEAPONLIST, iItem, iPlayer);
	return HAM_IGNORED;
}

public MsgHook_WeaponList(const iMsgID, const iMsgDest, const iMsgEntity)
{
	static arrWeaponListData[8];
	
	if (!iMsgEntity)
	{
		new szWeaponName[32];
		get_msg_arg_string(1, szWeaponName, charsmax(szWeaponName));
		
		if (!strcmp(szWeaponName, WEAPON_REFERANCE))
		{
			for (new i, a = sizeof arrWeaponListData; i < a; i++)
			{
				arrWeaponListData[i] = get_msg_arg_int(i + 2);
			}
		}
	}
	else
	{
		if (!IsCustomItem(iMsgDest) && pev(iMsgDest, pev_impulse))
		{
			return;
		}
		
		MESSAGE_BEGIN(MSG_ONE, iMsgID, {0.0, 0.0, 0.0}, iMsgEntity);
		WRITE_STRING(IsCustomItem(iMsgDest) ? WEAPON_NAME : WEAPON_REFERANCE);
		
		for (new i, a = sizeof arrWeaponListData; i < a; i++)
		{
			WRITE_BYTE(arrWeaponListData[i]);
		}
		
		MESSAGE_END();
	}
}

//**********************************************
//* Weaponbox world model.                     *
//**********************************************

public HamHook_Weaponbox_Spawn_Post(const iWeaponBox)
{
	if (IsValidPev(iWeaponBox))
	{
		state (IsValidPev(pev(iWeaponBox, pev_owner))) WeaponBox: Enabled;
	}
	
	return HAM_IGNORED;
}

public FakeMeta_SetModel(const iEntity) <WeaponBox: Enabled>
{
	state WeaponBox: Disabled;
	
	if (!IsValidPev(iEntity))
	{
		return FMRES_IGNORED;
	}
	
	#define MAX_ITEM_TYPES	6
	
	for (new i, iItem; i < MAX_ITEM_TYPES; i++)
	{
		iItem = get_pdata_cbase(iEntity, m_rgpPlayerItems_CWeaponBox + i, extra_offset_weapon);
		
		if (IsValidPev(iItem) && IsCustomItem(iItem))
		{
			SET_MODEL(iEntity, MODEL_WORLD);
			set_pev(iEntity, pev_body, 9);
			return FMRES_SUPERCEDE;
		}
	}
	
	return FMRES_IGNORED;
}

public FakeMeta_SetModel()	</* Empty statement */>	{ /*  Fallback  */ return FMRES_IGNORED; }
public FakeMeta_SetModel() 	< WeaponBox: Disabled >	{ /* Do nothing */ return FMRES_IGNORED; }

//**********************************************
//* Ammo Inventory.                            *
//**********************************************

PrimaryAmmoIndex(const iItem)
{
	return get_pdata_int(iItem, m_iPrimaryAmmoType, extra_offset_weapon);
}

GetAmmoInventory(const iPlayer, const iAmmoIndex)
{
	if (iAmmoIndex == -1)
	{
		return -1;
	}

	return get_pdata_int(iPlayer, m_rgAmmo_CBasePlayer + iAmmoIndex, extra_offset_player);
}

SetAmmoInventory(const iPlayer, const iAmmoIndex, const iAmount)
{
	if (iAmmoIndex == -1)
	{
		return 0;
	}

	set_pdata_int(iPlayer, m_rgAmmo_CBasePlayer + iAmmoIndex, iAmount, extra_offset_player);
	return 1;
}

stock Player_SetAnimation(const iPlayer, const szAnim[])
{
	if(!is_user_alive(iPlayer))
		return;
		
   #define ACT_RANGE_ATTACK1   28
   
   // Linux extra offsets
   #define extra_offset_animating   4
   
   // CBaseAnimating
   #define m_flFrameRate      36
   #define m_flGroundSpeed      37
   #define m_flLastEventCheck   38
   #define m_fSequenceFinished   39
   #define m_fSequenceLoops   40
   
   // CBaseMonster
   #define m_Activity      73
   #define m_IdealActivity      74
   
   // CBasePlayer
   #define m_flLastAttackTime   220
   
   new iAnimDesired, Float: flFrameRate, Float: flGroundSpeed, bool: bLoops;
      
   if ((iAnimDesired = lookup_sequence(iPlayer, szAnim, flFrameRate, bLoops, flGroundSpeed)) == -1)
   {
      iAnimDesired = 0;
   }
   
   new Float: flGametime = get_gametime();

   set_pev(iPlayer, pev_frame, 0.0);
   set_pev(iPlayer, pev_framerate, 1.0);
   set_pev(iPlayer, pev_animtime, flGametime );
   set_pev(iPlayer, pev_sequence, iAnimDesired);
   
   set_pdata_int(iPlayer, m_fSequenceLoops, bLoops, extra_offset_animating);
   set_pdata_int(iPlayer, m_fSequenceFinished, 0, extra_offset_animating);
   
   set_pdata_float(iPlayer, m_flFrameRate, flFrameRate, extra_offset_animating);
   set_pdata_float(iPlayer, m_flGroundSpeed, flGroundSpeed, extra_offset_animating);
   set_pdata_float(iPlayer, m_flLastEventCheck, flGametime , extra_offset_animating);
   
   set_pdata_int(iPlayer, m_Activity, ACT_RANGE_ATTACK1, extra_offset_player);
   set_pdata_int(iPlayer, m_IdealActivity, ACT_RANGE_ATTACK1, extra_offset_player);   
   set_pdata_float(iPlayer, m_flLastAttackTime, flGametime , extra_offset_player);
}

stock fm_get_aim_vector(index, forward_distance, Float:forward_origin[3], Float:view_end_origin[3])
{
	new Float:start[3], Float:view_ofs[3];
	pev(index, pev_origin, start);
	pev(index, pev_view_ofs, view_ofs);
	xs_vec_add(start, view_ofs, start);
	
	new Float:vector[3], Float:temp[3];
	pev(index, pev_v_angle, vector);
	engfunc(EngFunc_MakeVectors, vector);
	global_get(glb_v_forward, vector);
	xs_vec_mul_scalar(vector, float(forward_distance), temp);
	xs_vec_add(start, temp, forward_origin);
	
	new Float:dest[3];
	xs_vec_mul_scalar(vector, 9999.0, temp);
	xs_vec_add(start, temp, dest);
	engfunc(EngFunc_TraceLine, start, dest, 0, index, 0);
	get_tr2(0, TR_vecEndPos, view_end_origin);
	
	return true;
}

